unit Config;

interface

uses
  Interfaces,
  MainForm;

const
  ApplicationTitle = 'Deskew GUI';
  WebLink = 'http://galfar.vevb.net/deskew/';
  LogoImageResName = '';
  ShowDeskewExeOption = True;

procedure AfterMainFormCreation(MainForm: TFormMain);

implementation

procedure AfterMainFormCreation(MainForm: TFormMain);
begin
end;

end.

